''' Class containing OpenCL variables, such as context and que, along with methods that perform operations that are dependent on 
OpenCL. Input to the methods are geometry data'''

import numpy as np
import pyopencl as cl
import os

class Computation:
	def __init__(self, workPath):
		platforms = cl.get_platforms()
		my_CPU_devices  = platforms[0].get_devices(device_type = cl.device_type.CPU)
		my_GPU_devices  = platforms[0].get_devices(device_type = cl.device_type.GPU)

		os.environ['PYOPENCL_COMPILER_OUTPUT'] = '1'

		self.cntxt = cl.Context(devices=my_CPU_devices)
		self.queue = cl.CommandQueue(self.cntxt)

		# Load kernel file
		kernelFile = open(workPath+'kernels.cl', 'r')
		kernelStr  = "".join(kernelFile.readlines())
		kernelFile.close()

		# build kernel programs
		self.kernels = cl.Program(self.cntxt, kernelStr).build()


	def influenceMatrixSourceNewmann(self, meshData):
		A = np.zeros( (meshData.nrPoly, meshData.nrPoly) )

		a     = np.zeros(meshData.nrPoly, dtype = np.float32)
		aBuff = cl.Buffer(self.cntxt, cl.mem_flags.WRITE_ONLY, a.nbytes)

		for i in range(meshData.nrPoly):
			p = meshData.ctrlPoints[i]
			n = meshData.n[i]

			launch = self.kernels.sourceInfluenceNewmann(self.queue, 
												 		 (meshData.nrPoly,),
												 		 (1,),
												 		 p,
												 		 n,
												 		 meshData.verticesBuff,
												 		 meshData.ctrlPointsBuff,
												 		 meshData.lBuff,
												 		 meshData.mBuff,
												 		 meshData.nBuff,
												 		 meshData.topologyBuff,
												 		 aBuff)
			launch.wait()

			cl.enqueue_read_buffer(self.queue, aBuff, a).wait()

			A[i] = a

		return A

	def influenceMatrixSourceDirichlet(self, meshDataList):
		nrObjects = len(meshDataList)

		# Get overview of system size
		systemSize = 0
		for i in range(nrObjects):
			systemSize += meshDataList[i].nrPoly

		# Initialize influence matrix
		A = np.zeros( (systemSize, systemSize) ) 

		iRow = 0 # Row counter

		for objectNrRow in range(nrObjects): 
			meshDataRow = meshDataList[objectNrRow] # The object where the control point is evaluated 

			for i in range(meshDataRow.nrPoly):
				p = meshDataRow.ctrlPoints[i]
				aRow = np.array([])

				for objectNrCol in range(nrObjects):
					meshDataCol = meshDataList[objectNrCol]

					a     = np.zeros(meshDataCol.nrPoly, dtype = np.float32)
					aBuff = cl.Buffer(self.cntxt, cl.mem_flags.WRITE_ONLY, a.nbytes)

					launch = self.kernels.sourceInfluenceDirichlet(self.queue, 
												 		  		   (meshDataCol.nrPoly,),
												 		  		   (1,),
												 		  		   p,
												 		  		   meshDataCol.verticesBuff,
												 		  		   meshDataCol.ctrlPointsBuff,
												 		  		   meshDataCol.lBuff,
												 		  		   meshDataCol.mBuff,
												 		  		   meshDataCol.nBuff,
												 		  		   meshDataCol.topologyBuff,
												 		  		   aBuff)
					launch.wait()

					cl.enqueue_read_buffer(self.queue, aBuff, a).wait()

					aRow = np.append(aRow, a) # Add result to end of list

				A[iRow] = aRow

				iRow += 1

		return A

	def influenceMatrixDoubletNewmann(self, meshData):
		A = np.zeros( (meshData.nrPoly, meshData.nrPoly) )

		a     = np.zeros(meshData.nrPoly, dtype = np.float32)
		aBuff = cl.Buffer(self.cntxt, cl.mem_flags.WRITE_ONLY, a.nbytes)

		for i in range(meshData.nrPoly):
			p = meshData.ctrlPoints[i]
			n = meshData.n[i]

			launch = self.kernels.doubletInfluenceNewmann(self.queue, 
												 		 (meshData.nrPoly,),
												 		 (1,),
												 		 p,
												 		 n,
												 		 meshData.verticesBuff,
												 		 meshData.ctrlPointsBuff,
												 		 meshData.lBuff,
												 		 meshData.mBuff,
												 		 meshData.nBuff,
												 		 meshData.topologyBuff,
												 		 aBuff)
			launch.wait()

			cl.enqueue_read_buffer(self.queue, aBuff, a).wait()

			A[i] = a

		return A

	def influenceMatrixDoubletDirichlet(self, meshDataList):
		nrObjects = len(meshDataList)

		# Get overview of system size
		systemSize = 0
		for i in range(nrObjects):
			systemSize += meshDataList[i].nrPoly

		# Initialize influence matrix
		A = np.zeros( (systemSize, systemSize) ) 

		iRow = 0 # Row counter

		for objectNrRow in range(nrObjects): 
			meshDataRow = meshDataList[objectNrRow] # The object where the control point is evaluated 

			for i in range(meshDataRow.nrPoly):
				p = meshDataRow.ctrlPoints[i]
				aRow = np.array([])

				for objectNrCol in range(nrObjects):
					meshDataCol = meshDataList[objectNrCol]

					a     = np.zeros(meshDataCol.nrPoly, dtype = np.float32)
					aBuff = cl.Buffer(self.cntxt, cl.mem_flags.WRITE_ONLY, a.nbytes)

					launch = self.kernels.doubletInfluenceDirichlet(self.queue, 
												 		  		    (meshDataCol.nrPoly,),
												 		  		    (1,),
												 		  		    p,
												 		  		    meshDataCol.verticesBuff,
												 		  		    meshDataCol.ctrlPointsBuff,
												 		  		    meshDataCol.lBuff,
												 		  		    meshDataCol.mBuff,
												 		  		    meshDataCol.nBuff,
												 		  		    meshDataCol.topologyBuff,
												 		  		    aBuff)
					launch.wait()

					cl.enqueue_read_buffer(self.queue, aBuff, a).wait()

					aRow = np.append(aRow, a) # Add result to end of list

				A[iRow] = aRow

				iRow += 1

		return A

	def influenceFromStripsDirichlet(self, A, wingDataList, wakeDataList):

		nrObjects = len(wingDataList)

		iRow = 0

		for objectNrRow in range(nrObjects):
			wingDataRow = wingDataList[objectNrRow]
			wakeDataRow = wakeDataList[objectNrRow]

			for i in range(wingDataRow.nrPoly):
				p = wingDataRow.ctrlPoints[i]

				indexAdjust = 0

				for objectNrStrip in range(nrObjects):
					wingDataStrip = wingDataList[objectNrStrip]
					wakeDataStrip = wakeDataList[objectNrStrip]

					a = np.zeros( wakeDataStrip.nrPolyPrStrip, dtype = np.float32 )
					aBuff = cl.Buffer(self.cntxt, cl.mem_flags.WRITE_ONLY, a.nbytes)

					for j in range(wakeDataStrip.nrStrips):
					
						polyInStripBuff = cl.Buffer(self.cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = wakeDataStrip.polyInStrips[j])

						launch = self.kernels.stripInfluenceDirichlet(self.queue,
															  (wakeDataStrip.nrPolyPrStrip,),
															  (1,),
															  p,
															  wakeDataStrip.verticesBuff,
															  wakeDataStrip.ctrlPointsBuff,
															  wakeDataStrip.lBuff,
															  wakeDataStrip.mBuff,
															  wakeDataStrip.nBuff,
															  wakeDataStrip.topologyBuff,
															  polyInStripBuff,
															  aBuff)
						launch.wait()

						cl.enqueue_read_buffer(self.queue, aBuff, a).wait()

						trailingIndex1 = indexAdjust + wingDataStrip.polyInStrips[j][0]
						trailingIndex2 = indexAdjust + wingDataStrip.polyInStrips[j][-1]

						asum = a.sum()

						A[iRow][trailingIndex1] += asum
						A[iRow][trailingIndex2] -= asum

					indexAdjust += wingDataStrip.nrPoly

				iRow += 1

	def velocityAtPoint(self, p, wingDataList, wakeDataList, U):
		nrObjects = len(wingDataList)

		u = np.zeros(3) # Return value

		for i in range(nrObjects):
			wingData = wingDataList[i]
			wakeData = wakeDataList[i]

			aWing = np.zeros( (wingData.nrPoly, 4), dtype = np.float32 ) # Store values from wing
			aWingBuff = cl.Buffer(self.cntxt, cl.mem_flags.WRITE_ONLY, aWing.nbytes)

			aWake = np.zeros( (wakeData.nrPoly, 4), dtype = np.float32 ) # Store values from wake
			aWakeBuff = cl.Buffer(self.cntxt, cl.mem_flags.WRITE_ONLY, aWake.nbytes)

			# Find contribution from wing sources
			launch = self.kernels.velocitySource(self.queue, 
											 [wingData.nrPoly,],
											 None,
											 p,
											 wingData.verticesBuff,
											 wingData.ctrlPointsBuff,
											 wingData.lBuff,
											 wingData.mBuff,
											 wingData.nBuff,
											 wingData.topologyBuff,
											 wingData.sigmaBuff,
											 aWingBuff)
			launch.wait()

			cl.enqueue_read_buffer(self.queue, aWingBuff, aWing).wait()

			asum = aWing.sum(axis=0)

			u[0] += asum[0]
			u[1] += asum[1] 
			u[2] += asum[2] 

			# Find contribution from wing doublets
			launch = self.kernels.velocityDoublet(self.queue, 
											  [wingData.nrPoly,],
											  None,
											  p,
											  wingData.verticesBuff,
											  wingData.ctrlPointsBuff,
											  wingData.lBuff,
											  wingData.mBuff,
											  wingData.nBuff,
											  wingData.topologyBuff,
											  wingData.gammaBuff,
											  aWingBuff)
			launch.wait()

			cl.enqueue_read_buffer(self.queue, aWingBuff, aWing).wait()

			asum = aWing.sum(axis=0)

			u[0] += asum[0]
			u[1] += asum[1] 
			u[2] += asum[2] 

			# Find contribution from wake
			launch = self.kernels.velocityDoublet(self.queue, 
											  [wakeData.nrPoly,],
											  None,
											  p,
											  wakeData.verticesBuff,
											  wakeData.ctrlPointsBuff,
											  wakeData.lBuff,
											  wakeData.mBuff,
											  wakeData.nBuff,
											  wakeData.topologyBuff,
											  wakeData.gammaBuff,
											  aWakeBuff)
			launch.wait()

			cl.enqueue_read_buffer(self.queue, aWakeBuff, aWake).wait()

			asum = aWake.sum(axis=0)

			u[0] += asum[0]
			u[1] += asum[1] 
			u[2] += asum[2]

		# Free stream velocity
		u[0] += U[0]
		u[1] += U[1] 
		u[2] += U[2]

		return u

	def inducedVelocityAtPoint(self, p, wakeDataList, wingDataList, wing):
		nrObjects = len(wakeDataList)

		u = np.zeros(3) # Return value

		for i in range(nrObjects):
			wakeData = wakeDataList[i]

			aWake = np.zeros( (wakeData.nrPoly, 4), dtype = np.float32 ) # Store values from wake
			aWakeBuff = cl.Buffer(self.cntxt, cl.mem_flags.WRITE_ONLY, aWake.nbytes)

			wingData = wingDataList[i]

			if wingData != wing:
				aWing = np.zeros( (wingData.nrPoly, 4), dtype = np.float32 ) # Store values from wing
				aWingBuff = cl.Buffer(self.cntxt, cl.mem_flags.WRITE_ONLY, aWing.nbytes)

			

				# Find contribution from wing sources
				launch = self.kernels.velocitySource(self.queue, 
												 [wingData.nrPoly,],
												 None,
												 p,
												 wingData.verticesBuff,
												 wingData.ctrlPointsBuff,
												 wingData.lBuff,
												 wingData.mBuff,
												 wingData.nBuff,
												 wingData.topologyBuff,
												 wingData.sigmaBuff,
												 aWingBuff)
				launch.wait()

				cl.enqueue_read_buffer(self.queue, aWingBuff, aWing).wait()

				asum = aWing.sum(axis=0)

				u[0] += asum[0]
				u[1] += asum[1] 
				u[2] += asum[2] 

				# Find contribution from wing doublets
				launch = self.kernels.velocityDoublet(self.queue, 
												  [wingData.nrPoly,],
												  None,
												  p,
												  wingData.verticesBuff,
												  wingData.ctrlPointsBuff,
												  wingData.lBuff,
												  wingData.mBuff,
												  wingData.nBuff,
												  wingData.topologyBuff,
												  wingData.gammaBuff,
												  aWingBuff)
				launch.wait()

				cl.enqueue_read_buffer(self.queue, aWingBuff, aWing).wait()

				asum = aWing.sum(axis=0)

				u[0] += asum[0]
				u[1] += asum[1] 
				u[2] += asum[2] 

			# Find contribution from wake
			launch = self.kernels.velocityDoublet(self.queue, 
											  [wakeData.nrPoly,],
											  None,
											  p,
											  wakeData.verticesBuff,
											  wakeData.ctrlPointsBuff,
											  wakeData.lBuff,
											  wakeData.mBuff,
											  wakeData.nBuff,
											  wakeData.topologyBuff,
											  wakeData.gammaBuff,
											  aWakeBuff)
			launch.wait()

			cl.enqueue_read_buffer(self.queue, aWakeBuff, aWake).wait()

			asum = aWake.sum(axis=0)

			u[0] += asum[0]
			u[1] += asum[1] 
			u[2] += asum[2]

		return u


	def rightSideSigma(self, meshDataList, U):
		nrObjects = len(meshDataList)

		systemSize = 0

		for objectNr in range(nrObjects):
			systemSize += meshDataList[objectNr].nrPoly

		sigma = np.zeros(systemSize)

		iRow = 0

		for objectNr in range(nrObjects):	
			meshData = meshDataList[objectNr]

			for i in range(meshData.nrPoly):
				n = meshData.n[i]

				sigma[iRow] = -(n[0]*U[0] + n[1]*U[1] + n[2]*U[2])/(4*np.pi)

				iRow += 1

		return sigma

	def deformWake(self, wake, wingDataList, wakeDataList, U):

		Uinf = np.sqrt( U[0]**2 + U[1]**2 + U[2]**2 )

		for i in range(wake.nrPolyPrStrip):
			for j in range(wake.nrStrips+1):
				vertIndex1 = (wake.nrPolyPrStrip+1)*(1 + j) - i - 1
				vertIndex2 = (wake.nrPolyPrStrip+1)*(1 + j) - i - 2
				
				p1 = wake.vertices[vertIndex1]
				p2 = wake.vertices[vertIndex2]

				p11 = 0.75*p1 + 0.25*p2
				p22 = 0.25*p1 + 0.75*p2

				dist = p2 - p1
				dx = np.sqrt( dist[0]**2 + dist[1]**2 + dist[2]**2 ) 

				dt = dx/Uinf

				u1 = self.velocityAtPoint(p11, wingDataList, wakeDataList, U)
				u2 = self.velocityAtPoint(p22, wingDataList, wakeDataList, U)

				u = 0.5*(u1 + u2)

				dp = np.zeros(3)
				dp[0] = u[0]*dt
				dp[1] = u[1]*dt
				dp[2] = u[2]*dt
				
				wake.object.data.vertices[vertIndex2].co[0] = wake.object.data.vertices[vertIndex1].co[0] + dp[0]
				wake.object.data.vertices[vertIndex2].co[1] = wake.object.data.vertices[vertIndex1].co[1] + dp[1]
				wake.object.data.vertices[vertIndex2].co[2] = wake.object.data.vertices[vertIndex1].co[2] + dp[2]

			wake.object.data.update()
			wake.updateData(self.cntxt)

	def forceCalculation(self, wake, wing, wakeDataList, wingDataList, U):
		Uinf = np.sqrt( U[0]**2 + U[1]**2 + U[2]**2 )

		Force = np.zeros(3)

		for i in range(wake.nrStrips):
			polyIndex = wake.polyInStrips[i][-1]

			vertIndex1 = (wake.nrPolyPrStrip+1)*(1 + i) - 1
			vertIndex2 = (wake.nrPolyPrStrip+1)*(2 + i) - 1

			p1 = wake.vertices[vertIndex1]
			p2 = wake.vertices[vertIndex2]

			dist = np.sqrt( (p2[0] - p1[0])**2 + (p2[1] - p1[1])**2 + (p2[2] - p1[2])**2)

			gammaVec = (p2 - p1)/dist

			p = 0.5*(p1 + p2)  

			u = self.inducedVelocityAtPoint(p, wakeDataList, wingDataList, wing)

			gamma = wake.gamma[polyIndex]*gammaVec

			forceFactor = 4*np.pi*dist*2/Uinf**2

			Force[0] +=  forceFactor*((U[1] + u[1])*gamma[2] - (U[2] + u[2])*gamma[1])
			Force[1] +=  forceFactor*((U[2] + u[2])*gamma[0] - (U[0] + u[0])*gamma[2])
			Force[2] +=  forceFactor*((U[1] + u[1])*gamma[0] - (U[0] + u[0])*gamma[1])

		return Force




		
				







