import bpy
import numpy as np
import mathutils
import ColorMaps
import pyopencl as cl
import os

class MeshData:

	def __init__(self, objectName):
		''' Class that contains the blender data, and data formatet to be sent to computations'''
		
		''' Blender data '''
		self.object = bpy.data.objects[objectName]

		# Create rotation matrix for transforming normals to global coordinate system
		euler_x = self.object.rotation_euler[0]
		euler_y = self.object.rotation_euler[1]
		euler_z = self.object.rotation_euler[2]
		Rx = mathutils.Matrix.Rotation(euler_x, 4, 'X')
		Ry = mathutils.Matrix.Rotation(euler_y, 4, 'Y')
		Rz = mathutils.Matrix.Rotation(euler_z, 4, 'Z')

		self.matrix_rotation = Rz * Ry * Rx
		self.matrix_world    = self.object.matrix_world

		self.polygons = self.object.data.polygons

		''' OpenCL formatet data '''
		self.nrPoly = len(self.polygons)
		self.nrVert = len(self.object.data.vertices)

		self.vertices   = np.zeros((self.nrVert, 4), dtype = np.float32) # Global coordinates
		self.ctrlPoints = np.zeros((self.nrPoly, 4), dtype = np.float32) # Global coordinates
		self.topology   = np.zeros((self.nrPoly, 4), dtype = np.int32  ) # Topology in opencl formate

		# Local coordinate system
		self.l = np.zeros((self.nrPoly, 4), dtype = np.float32)
		self.m = np.zeros((self.nrPoly, 4), dtype = np.float32)
		self.n = np.zeros((self.nrPoly, 4), dtype = np.float32)

		# Go through all vertices
		for i in range(self.nrVert):
			p = self.matrix_world * self.object.data.vertices[i].co # Transform to global coordinates

			self.vertices[i][0] = np.float32(p[0])
			self.vertices[i][1] = np.float32(p[1])
			self.vertices[i][2] = np.float32(p[2])

		# Go through all polygons
		for i in range(self.nrPoly):
			n = self.matrix_rotation * self.polygons[i].normal # Transform to global coordinates
			p = self.matrix_world    * self.polygons[i].center

			i1 = self.polygons[i].vertices[0]
			i2 = self.polygons[i].vertices[1]
			i3 = self.polygons[i].vertices[2]
			i4 = self.polygons[i].vertices[3]

			p1 = self.vertices[i1]
			p2 = self.vertices[i2]
			p3 = self.vertices[i3]
			p4 = self.vertices[i4]

			l = (p1 + p2 - p3 - p4)
			l = l/np.sqrt( l[0]**2 + l[1]**2 + l[2]**2 )

			m = np.zeros(3)
			m[0] = n[1]*l[2] - l[1]*n[2]
			m[1] = l[0]*n[2] - n[0]*l[2]
			m[2] = n[0]*l[1] - l[0]*n[1]

			m = m/np.sqrt( m[0]**2 + m[1]**2 + m[2]**2 )

			for j in range(3):
				self.ctrlPoints[i][j] = np.float32(p[j])

				self.l[i][j] = np.float32(l[j])
				self.m[i][j] = np.float32(m[j])
				self.n[i][j] = np.float32(n[j])

			for j in range(4):
				self.topology[i][j] = np.int32(self.polygons[i].vertices[j])


		''' Initialize variables for later use '''
		self.velocity = np.zeros( (self.nrPoly, 4) )
		self.pressure = np.zeros(  self.nrPoly )
		self.Force    = np.zeros(3)

		self.vertVelocity = np.zeros( (self.nrVert, 4) )
		self.vertPressure = np.zeros(  self.nrVert )

		self.nrStrips = 0
		self.nrPolyPrStrip = 0

	def findStrips(self, nrStrips, nrPolyPrStrip = 0):
		''' Divide mesh into strips, based on assumption that it os from NURBS object in blender'''

		self.nrStrips = nrStrips

		if nrPolyPrStrip == 0:
			self.nrPolyPrStrip = np.int(self.nrPoly/nrStrips)
		else:
			self.nrPolyPrStrip = nrPolyPrStrip

		self.polyInStrips = np.zeros( (self.nrStrips, self.nrPolyPrStrip), dtype = np.int32 ) # Indices of polygons in strips
		self.stripForPoly = np.zeros(self.nrPoly) # Which strip each polygon belongs to

		iPoly = 0
		for i in range(nrStrips):
			for j in range(self.nrPolyPrStrip):
				self.polyInStrips[i][j] = np.int32(iPoly)
				
				self.stripForPoly[iPoly] = i

				iPoly += 1

		# Creating colormap to seperate out the strips
		colorMaps = self.object.data.vertex_colors

		createColorMap = True
		colorMapName = 'strips'

		for i in range(len(colorMaps)):
			if colorMaps[i].name == colorMapName:
				createColorMap = False

		if createColorMap:
			colorMaps.new(colorMapName)

		iColor = 0
		for iPoly in range(self.nrPoly):
			strip = self.stripForPoly[iPoly]

			rgb = ColorMaps.jet(strip, nrStrips-1, 0)

			for iTop in range(4):
				colorMaps[colorMapName].data[iColor].color = rgb

				iColor += 1

	def setOpenCL(self, cntxt):
		self.topologyBuff   = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.topology)
		self.verticesBuff   = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.vertices)
		self.lBuff          = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.l)
		self.mBuff          = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.m)
		self.nBuff          = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.n)
		self.ctrlPointsBuff = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.ctrlPoints)

	def updateData(self, cntxt):
		# Go through all vertices
		for i in range(self.nrVert):
			p = self.matrix_world * self.object.data.vertices[i].co # Transform to global coordinates

			self.vertices[i][0] = np.float32(p[0])
			self.vertices[i][1] = np.float32(p[1])
			self.vertices[i][2] = np.float32(p[2])

		# Go through all polygons
		for i in range(self.nrPoly):
			n = self.matrix_rotation * self.polygons[i].normal # Transform to global coordinates
			p = self.matrix_world    * self.polygons[i].center

			i1 = self.polygons[i].vertices[0]
			i2 = self.polygons[i].vertices[1]
			i3 = self.polygons[i].vertices[2]
			i4 = self.polygons[i].vertices[3]

			p1 = self.vertices[i1]
			p2 = self.vertices[i2]
			p3 = self.vertices[i3]
			p4 = self.vertices[i4]

			l = (p1 + p2 - p3 - p4)
			l = l/np.sqrt( l[0]**2 + l[1]**2 + l[2]**2 )

			m = np.zeros(3)
			m[0] = n[1]*l[2] - l[1]*n[2]
			m[1] = l[0]*n[2] - n[0]*l[2]
			m[2] = n[0]*l[1] - l[0]*n[1]

			m = m/np.sqrt( m[0]**2 + m[1]**2 + m[2]**2 )

			for j in range(3):
				self.ctrlPoints[i][j] = np.float32(p[j])

				self.l[i][j] = np.float32(l[j])
				self.m[i][j] = np.float32(m[j])
				self.n[i][j] = np.float32(n[j])

		# Update buffers
		self.verticesBuff   = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.vertices)
		self.lBuff          = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.l)
		self.mBuff          = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.m)
		self.nBuff          = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.n)
		self.ctrlPointsBuff = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.ctrlPoints)

	def setGamma(self, gamma, cntxt):
		self.gamma = np.float32(gamma)

		self.gammaBuff = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.gamma)

	def setSigma(self, sigma, cntxt):
		self.sigma = np.float32(sigma)

		self.sigmaBuff = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.sigma)

	def setGammaWake(self, wingData, cntxt):
		self.gamma = np.zeros(self.nrPoly, dtype = np.float32)

		for i in range(self.nrStrips):
			trailingIndex1 = wingData.polyInStrips[i][0]
			trailingIndex2 = wingData.polyInStrips[i][-1]

			g = wingData.gamma[trailingIndex1] - wingData.gamma[trailingIndex2]

			for j in range(self.nrPolyPrStrip):
				index = self.polyInStrips[i][j]
				self.gamma[index] = g

		self.gammaBuff = cl.Buffer(cntxt, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf = self.gamma)

	def postProcessingDirichlet(self, U):

		Uinf = U[0]**2 + U[1]**2 + U[2]**2
		# Go through all polygons once
		for i in range(self.nrStrips):
			for j in range(self.nrPolyPrStrip):

				# Current polygon
				i0 = self.polyInStrips[i][j]
				p0 = self.ctrlPoints[i0]

				if (i==0):
					iN = self.polyInStrips[i+1][j]

					pN = self.ctrlPoints[iN]

					gamma_N2 = self.gamma[iN]
					gamma_N1 = self.gamma[i0]

					L_m = np.sqrt( (pN[0] - p0[0])**2 + (pN[1] - p0[1])**2 + (pN[2] - p0[2])**2 )
				elif (i == self.nrStrips - 1):
					iS = self.polyInStrips[i-1][j]

					pS = self.ctrlPoints[iS]

					gamma_N2 = self.gamma[i0]
					gamma_N1 = self.gamma[iS]

					L_m = np.sqrt( (p0[0] - pS[0])**2 + (p0[1] - pS[1])**2 + (p0[2] - pS[2])**2 )
				else:
					iN = self.polyInStrips[i+1][j]
					iS = self.polyInStrips[i-1][j]

					pN = self.ctrlPoints[iN]
					pS = self.ctrlPoints[iS]

					gamma_N2 = self.gamma[iN]
					gamma_N1 = self.gamma[iS]

					L_m = np.sqrt( (pN[0] - p0[0])**2 + (pN[1] - p0[1])**2 + (pN[2] - p0[2])**2 ) + np.sqrt( (p0[0] - pS[0])**2 + (p0[1] - pS[1])**2 + (p0[2] - pS[2])**2 )

				if (j==0):
					iE = self.polyInStrips[i][j+1]

					pE = self.ctrlPoints[iE]

					gamma_E2 = self.gamma[iE]
					gamma_0  = self.gamma[i0]

					SA = np.sqrt( (pE[0] - p0[0])**2 + (pE[1] - p0[1])**2 + (pE[2] - p0[2])**2 )
					SB = 0

					DA = (gamma_E2 - gamma_0)/SA
					DB = 0

					L_l = np.sqrt( (pE[0] - p0[0])**2 + (pE[1] - p0[1])**2 + (pE[2] - p0[2])**2 )
				elif (j==self.nrPolyPrStrip-1):
					iW = self.polyInStrips[i][j-1]

					pW = self.ctrlPoints[iW]

					gamma_0  = self.gamma[i0]
					gamma_E1 = self.gamma[iW]

					SA = 0
					SB = np.sqrt( (pW[0] - p0[0])**2 + (pW[1] - p0[1])**2 + (pW[2] - p0[2])**2 )
					DA = 0
					DB = (gamma_E1 - gamma_0)/SB
				else:
					iE = self.polyInStrips[i][j+1]
					iW = self.polyInStrips[i][j-1]

					pE = self.ctrlPoints[iE]
					pW = self.ctrlPoints[iW]

					gamma_E2 = self.gamma[iE]
					gamma_E1 = self.gamma[iW]
					gamma_0  = self.gamma[i0]

					SA = np.sqrt( (pE[0] - p0[0])**2 + (pE[1] - p0[1])**2 + (pE[2] - p0[2])**2 )
					SB = np.sqrt( (pW[0] - p0[0])**2 + (pW[1] - p0[1])**2 + (pW[2] - p0[2])**2 )

					DA = (gamma_E2 - gamma_0)/SA
					DB = (gamma_E1 - gamma_0)/SB


				u_l = -4*np.pi*( DA*SA - DB*SB )/(SA + SB)
				u_m = -4*np.pi*( gamma_N2 - gamma_N1 )/L_m
				u_n = 4*np.pi*self.sigma[i0]


				# Transformation to global coordinate system
				u = u_l*self.l[i0] + u_m*self.m[i0] + u_n*self.n[i0]

				self.velocity[i0][0] = U[0] + u[0]
				self.velocity[i0][1] = U[1] + u[1]
				self.velocity[i0][2] = U[2] + u[2]

				self.velocity[i0][3] = np.sqrt( self.velocity[i0][0]**2 + self.velocity[i0][1]**2 + self.velocity[i0][2]**2 )
				
				self.pressure[i0] = 1 - (self.velocity[i0][3]**2)/Uinf

				force = -self.pressure[i0]*self.n[i0]*self.object.data.polygons[i0].area

				self.Force[0] += force[0]
				self.Force[1] += force[1]
				self.Force[2] += force[2]

	def velocityMagnitude(self):
		for i in range(self.nrPoly):
			self.velocity[i][3] = np.sqrt( self.velocity[i][0]**2 + self.velocity[i][1]**2 + self.velocity[i][2]**2 )

	def vertVelocityMagnitude(self):
		for i in range(self.nrVert):
			self.vertVelocity[i][3] = np.sqrt( self.vertVelocity[i][0]**2 + self.vertVelocity[i][1]**2 + self.vertVelocity[i][2]**2 )


	def velocityColorMap(self):
		colorMaps = self.object.data.vertex_colors

		createColorMap = True
		colorMapName = 'velocity'

		for i in range(len(colorMaps)):
			if colorMaps[i].name == colorMapName:
				createColorMap = False

		if createColorMap:
			colorMaps.new(colorMapName)

		uMax = self.velocity[0][3]
		uMin = self.velocity[0][3]

		for i in range(self.nrPoly):
			if uMax < self.velocity[i][3]:
				uMax = self.velocity[i][3]
			if uMin > self.velocity[i][3]:
				uMin = self.velocity[i][3]

		print('Min velcoity magnitude:', uMin)
		print('Max velcoity magnitude:', uMax)

		iColor = 0
		for iPoly in range(self.nrPoly):
			u = self.velocity[iPoly][3]

			rgb = ColorMaps.jet(u, uMax, uMin)

			for iTop in range(4):
				colorMaps[colorMapName].data[iColor].color = rgb

				iColor += 1

	def vertVelocityColorMap(self):
		colorMaps = self.object.data.vertex_colors

		createColorMap = True
		colorMapName = 'vertVelocity'

		for i in range(len(colorMaps)):
			if colorMaps[i].name == colorMapName:
				createColorMap = False

		if createColorMap:
			colorMaps.new(colorMapName)

		uMax = self.vertVelocity[0][3]
		uMin = self.vertVelocity[0][3]

		for i in range(self.nrVert):
			if uMax < self.vertVelocity[i][3]:
				uMax = self.vertVelocity[i][3]
			if uMin > self.vertVelocity[i][3]:
				uMin = self.vertVelocity[i][3]

		print('Min vertVelcoity magnitude:', uMin)
		print('Max vertVelcoity magnitude:', uMax)

		iColor = 0
		for iPoly in range(self.nrPoly):

			for iTop in range(4):
				iVert = self.polygons[iPoly].vertices[iTop]

				u = self.vertVelocity[iVert][3]

				rgb = ColorMaps.jet(u, uMax, uMin)

				colorMaps[colorMapName].data[iColor].color = rgb

				iColor += 1

	def pressureColorMap(self):
		colorMaps = self.object.data.vertex_colors

		createColorMap = True
		colorMapName = 'pressure'

		for i in range(len(colorMaps)):
			if colorMaps[i].name == colorMapName:
				createColorMap = False

		if createColorMap:
			colorMaps.new(colorMapName)

		pMax = self.pressure[0]
		pMin = self.pressure[0]

		for i in range(self.nrPoly):
			if pMax < self.pressure[i]:
				pMax = self.pressure[i]
			if pMin > self.pressure[i]:
				pMin = self.pressure[i]

		print('Min pressure coefficient:', pMin)
		print('Max pressure coefficient:', pMax)

		iColor = 0
		for iPoly in range(self.nrPoly):
			p = self.pressure[iPoly]

			rgb = ColorMaps.jet(p, pMax, pMin)

			for iTop in range(4):
				colorMaps[colorMapName].data[iColor].color = rgb

				iColor += 1





